/* 
 * Copyright (c) 1995
 *       Whitehead Institute for Biomedical Research. All rights reserved.
 *
 * Released under terms specified in the Copyright file present in the top
 * directory of this distribution.
 */


#ifndef BASICS_H
#define BASICS_H
// $Header: /usr/local/db4/cvsroot/steve/lbback/src/basics.H,v 1.4 1995/02/13 17:49:06 steve Exp $

#ifndef __STDLIB_H
#include <stdlib.h>
#endif

#ifndef STREAMH
#include <stream.h>
#endif

// The ObjectStore compiler seems not to provide ## and # preprocessor
// operators.
#if defined(OBJECTSTORE)
#define CAT_ID2(N1,N2) N1/**/N2
#define MACRO_STRING(X) "X"
#else
#define CAT_ID2(N1,N2) N1##N2
#define MACRO_STRING(X) #X
#endif


// The next preprocessor conditional provides alternative #defines for
// database-related types.

// ============================================================================
#if defined(OBJECTSTORE)

#ifndef _OBJECTSTORE_H_
#include <objectstore/objectstore.H>
#endif

#ifndef _OS_COLLECTION_H_
#include <ostore/coll.hh>
#endif

#define P_SEGMENT os_segment
#define P_DATABASE os_database
#define P_DATABASE_ROOT os_database_root
#define P_TRANSACTION os_transaction

#define PNEW_ADDR(Address,Type,Args)       \
  new(os_segment::of(Address)) Type Args
#define PNEW_ARRAY_ADDR(Address,Type,Size) \
  new(os_segment::of(Address)) Type[Size]
#define PNEW(Segment,Type,Args)            new(Segment) Type Args
#define PNEW_ARRAY(Segment,Type,Size)      new(Segment) Type[Size]

// ============================================================================
#elif defined(TEXAS)

#ifndef TEXAS_HH
#include "texas.hh"
#endif

#define P_SEGMENT ObjStore
#define P_DATABASE PStore
#define P_DATABASE_ROOT void
#define P_TRANSACTION char

#define PNEW_ADDR(Address,Type,Args) \
  (obj_store_of(Address) == (ObjStore*)p_store ?  \
      pnew(p_store,Type) Args : new Type Args)
#define PNEW_ARRAY_ADDR(Address,Type,Size) \
  (obj_store_of(Address) == (ObjStore*)p_store ? \
     pnew_array(p_store,Type,Size) : new Type[Size])
#define PNEW(Segment,Type,Args) (p_store ? \
      pnew(p_store,Type) Args : new Type Args)
#define PNEW_ARRAY(Segment,Type,Size) (p_store ? \
     pnew_array(p_store,Type,Size) : new Type[Size])

// We use the following global, which contains the same value as QL_DD::db() to
// reduce dependency on class QL_DD.  p_store is 0 if all data is to be stored
// in main memory.
extern PStore *p_store;

#else // No database at all ===================================================

#define P_SEGMENT char
#define P_DATABASE char
#define P_DATABASE_ROOT void
#define P_TRANSACTION char

#define PNEW_ADDR(Segment,Type,Args)	    new Type Args
#define PNEW_ARRAY_ADDR(Segment,Type,Size)  new Type[Size]
#define PNEW(Segment,Type,Args)	            new Type Args
#define PNEW_ARRAY(Segment,Type,Size)       new Type[Size]

#endif


inline int min (const int x, const int y)
{
  return (x < y) ? x : y;
}


inline int max (const int x, const int y)
{
  return (x > y) ? x : y;
}


#if 0
inline int ceiling (const int x, const int y)
{
  return 1 + (x - 1)/y;
}


inline int floor (const int x, const int y)
{
  return x/y;
}
#endif

// We have a very very simple error-handling scheme.
// Error messages are stored as a char*, accessible
// through errmsg(), cleared with clear_error() and
// set with set_error()
//
extern char* _LAST_ERR;
inline char* errmsg () { return _LAST_ERR; }
inline void set_error (char* a) { _LAST_ERR = a;}
inline void clear_error () { _LAST_ERR = NULL; }

extern void die_and_dump_core();


// Use ABORT to create failstop (failsafe) functions in the face of
// programming errors.  I.e., if an input would violate the entry
// invariants of a function, test for that in ABORT.  ABORT prints an
// error message (the second argument) and calls abort() if its first
// argument is non-zero.
//
#define ABORT( condition, message ) \
if (condition) { \
 (cerr << "ABORT " << __FILE__ << ":" << __LINE__ \
       << " " << message << endl).flush(); \
 die_and_dump_core(); }


// The SLOT macro defines a protected class slot with type TYPE, name
// _NAME, and a public accessor function called NAME.  Members defined
// after SLOT are by default protected.
// 
#define SLOT(TYPE,NAME) public: TYPE NAME() const { return CAT_ID2(_,NAME); } \
                        protected: TYPE CAT_ID2(_,NAME)


#define MAX_ARITY 24  // Maximum arity of arguments in the query language.


#endif			// last line of file
