/* 
 * Copyright (c) 1995
 *       Whitehead Institute for Biomedical Research. All rights reserved.
 *
 * Released under terms specified in the Copyright file present in the top
 * directory of this distribution.
 */


#ifndef BUILTIN_H
#define BUILTIN_H
// $Header: /usr/local/db4/cvsroot/steve/lbback/src/builtin.H,v 1.4 1995/02/13 17:49:08 steve Exp $

#ifndef BASICS_H
#include "basics.H"
#endif

#ifndef VALUE_H
#include "value.H"
#endif

#ifndef QL_ERROR_H
#include "ql_error.H"
#endif

class QL_Iterator_State;
class QL_Term;

#define QL_FIRST_HEADER(NAME) \
QL_Iterator_State * NAME(QL_Term*t, int m, Value*in[], int n, Value*out[])

#define BUILTIN_FIRST_EXTERN(NAME) extern QL_FIRST_HEADER(NAME);

typedef QL_FIRST_HEADER((*Builtin_First_Fn));

#define QL_NEXT_HEADER(NAME) \
QL_Iterator_State * \
  NAME(QL_Term*t, QL_Iterator_State*s, int m, Value*in[], int n, Value*out[])

#define BUILTIN_NEXT_EXTERN(NAME) extern QL_NEXT_HEADER(NAME);


typedef QL_NEXT_HEADER((*Builtin_Next_Fn));


static inline void abort_m_n(int m, int m_need, int n, int n_need)
{
  ABORT(m != m_need, "Wrong args");
  ABORT(n != n_need, "Wrong args");
}

#define BUILTIN_FIRST_HEADER(FUNCTION_NAME) \
     QL_Iterator_State *FUNCTION_NAME \
     (QL_Term *t, int m, Value*in_tv[], int n, Value*out_tv[])


#define BUILTIN_NEXT_HEADER(FUNCTION_NAME) \
  QL_Iterator_State* FUNCTION_NAME(QL_Term *t, QL_Iterator_State *s, \
					  int n, Value*out_tv[])


#define BUILTIN_FIRST_DEFINITION(FN_NAME,M,N,IN_TYPES,TYPE_NUMBER,GC_IS_OK) \
  static BUILTIN_FIRST_HEADER(CAT_ID2(FN_NAME,_internal)); \
  QL_FIRST_HEADER(FN_NAME) \
  { \
  return inject_and_typecheck_first(t, m, in, n, out, \
                                    IN_TYPES, TYPE_NUMBER, M, N, GC_IS_OK, \
				    CAT_ID2(FN_NAME,_internal)); \
  } \
  BUILTIN_FIRST_HEADER(CAT_ID2(FN_NAME,_internal))


#define BUILTIN_NEXT_DEFINITION(FN_NAME,GC_IS_OK) \
  static BUILTIN_NEXT_HEADER(CAT_ID2(FN_NAME,_internal)); \
  QL_NEXT_HEADER(FN_NAME) \
  { if (in) m=0; return inject_next(t, s, n, out, GC_IS_OK, \
    CAT_ID2(FN_NAME,_internal)); } \
  BUILTIN_NEXT_HEADER(CAT_ID2(FN_NAME,_internal))

extern int coerce_to_int(const QL_Term *, const char *, Value*, int*);
// Return 1 on success, 0 on failure.
// Put the result in the last arg.
// Use the term and const char * for calling ql_eval_error.
// The const char * is prepended to the message.

extern QL_Iterator_State *
inject_and_typecheck_first(QL_Term*t, int m, Value*in[],
                           int n, Value*out[],
                           Value::tv_type_id in_types[],
                           int type_number,
			   int expected_m, int expected_n, int gc_is_ok,
                           BUILTIN_FIRST_HEADER((*first_fn)));

extern QL_Iterator_State *
inject_next(QL_Term*t, QL_Iterator_State *s,
	    int n, Value*out[], int gc_is_ok,
	    BUILTIN_NEXT_HEADER((*next_fn)));

extern QL_Term *tv2term(QL_Term *, Value *);
extern int check_arg_type(QL_Term *,
			  Value **, Value::tv_type_id, int arg_number);

extern int type_ok(QL_Term *t, Value **v, Value::tv_type_id type);
#endif
