/* 
 * Copyright (c) 1995
 *       Whitehead Institute for Biomedical Research. All rights reserved.
 *
 * Released under terms specified in the Copyright file present in the top
 * directory of this distribution.
 */


#ifndef QL_DD_H
#include "ql_dd.H"
#endif

#ifndef QL_DD_ENTRY_H
#include "ql_dd_entry.H"
#endif

#ifndef QL_TERM_H
#include "ql_term.H"
#endif

#ifndef QL_ARGS_H
#include "ql_args.H"
#endif

#ifndef BUILTIN3__H
#include "builtin3.H"
#endif

#ifndef LIST_H
#include "list.H"
#endif

#ifndef STEP_H
#include "step.H"
#endif

#ifndef MATERIAL_H
#include "material.H"
#endif

static Value::tv_type_id one_undefined[] = { Value::UNDEFINED };

static QL_DD_Entry *plus_entry=0, *minus_entry=0, *times_entry=0, *div_entry=0;

static Value *op2(QL_Term *t)
{
  // We are guaranteed exactly two arguments.
  QL_Args *a0 = t->args(), *a1 = a0->next();
  Value *v0 = a0->value(), *v1 = a1->value(), *v2;
  if (v0->tv_type() == Value::TERM) v0 = op2(v0->down_to_TERM());
  if (v1->tv_type() == Value::TERM) v1 = op2(v1->down_to_TERM());
  if (!v0 || !v1) return 0;
  QL_DD_Entry *x = t->dd_entry();
  if (x != plus_entry && x != minus_entry
      && x != times_entry && x != div_entry) {
    ql_eval_error(t, "Unrecognized operator, '", x->identifier(), "' in");
    return 0;
  }
  Value::tv_type_id t0 = v0->tv_type(), t1 = v1->tv_type();
  clear_error();
  if (t0 == Value::FLOAT || t1 == Value::FLOAT) {
    double d0, d1, d2;
    switch (t0) {
    case Value::FLOAT: d0 = (double) *v0->down_to_FLOAT(); break;
    case Value::INTEGER32: d0 = (int) *v0->down_to_INTEGER32(); break;
    case Value::INTEGER: d0 = (int) *v0->down_to_INTEGER(); break;
    default: ql_eval_error(t, "Argument 1 must be a number in"); return 0;
    }
    switch (t1) {
    case Value::FLOAT: d1 = (double) *v1->down_to_FLOAT(); break;
    case Value::INTEGER32: d1 = (int) *v1->down_to_INTEGER32(); break;
    case Value::INTEGER: d1 = (int) *v1->down_to_INTEGER(); break;
    default: ql_eval_error(t, "Argument 2 must be a number in"); return 0;
    }
    if (x == plus_entry) d2 = d0 + d1;
    else if (x == minus_entry) d2 = d0 - d1;
    else if (x == times_entry) d2 = d0 * d1;
    else if (x == div_entry) d2 = d0 / d1;
    v2 =  Float::make(d2);
  } else {
    long int i0, i1, i2;
    switch (t0) {
    case Value::INTEGER: i0 = (int) *v0->down_to_INTEGER(); break;
    case Value::INTEGER32: i0 = (int) *v0->down_to_INTEGER32(); break;
    default: ql_eval_error(t, "Argument 1 must be a number in"); return 0;
    }
    switch (t1) {
    case Value::INTEGER: i1 = (int) *v1->down_to_INTEGER(); break;
    case Value::INTEGER32: i1 = (int) *v1->down_to_INTEGER32(); break;
    default: ql_eval_error(t, "Argument 2 must be a number in"); return 0;
    }
    if (x == plus_entry) i2 = i0 + i1;
    else if (x == minus_entry) i2 = i0 - i1;
    else if (x == times_entry) i2 = i0 * i1;
    else if (x == div_entry) i2 = i0 / i1;
    v2 = Long_Int::make(i2);
  }
  if (errmsg()) {
    ql_eval_error(t, errmsg(), " in");
    return 0;
  }
  return v2;
}


BUILTIN_FIRST_DEFINITION(no_eval_pred,-1,-1,one_undefined,1,0)
{
  if (m || n || in_tv || out_tv);
  ql_eval_error(t,"It is not possible to evaluate");
  return 0;
}


BUILTIN_FIRST_DEFINITION(is_pred,1,1,one_undefined,1,1)
{
  if (t || m || n); // Suppress errors.
  out_tv[0] = 0;
  if (!plus_entry) {
    plus_entry = QL_DD::find_or_error("+");
    minus_entry =  QL_DD::find_or_error("-");
    times_entry =  QL_DD::find_or_error("*");
    div_entry =  QL_DD::find_or_error("/");
  }
  if (!plus_entry || !minus_entry || !times_entry || !div_entry) return 0;
  QL_Term *t0;
  switch (in_tv[0]->tv_type()) {
  case Value::TERM:
    t0 = in_tv[0]->down_to_TERM();
    if (2 != t0->args()->arity()) {
      ql_eval_error(t0, "Expect 2 arguments in");
      return 0;
    }
    out_tv[0] = op2(t0);
    break;
  default: out_tv[0] = in_tv[0]; break;
  }
  return out_tv[0] ? (QL_Iterator_State*)1 : 0;
}


static Value*
s_tag_values(const QL_Term*, const Value *v)
{
  const Step *s = v->down_to_STEP_POINTER();
  Tag *x = s->tag_array();
  int len = s->tag_number();
  const int naddl = 2;
  int i = 0;
  Value **pairs = new Value* [len + naddl], *pair[2], *ret;

  pair[0]
    = String::make(QL_DD::find_by_tag_index(Tag::step_kind())->identifier());
  pair[1] = s->find(Tag::step_kind());
  ABORT(!pair[1],"Tag::step_kind not found");
  pairs[i] = V_Collection::make(Value::LIST, 2, pair);
  i++;
  pair[0] = String::make(QL_DD::find_by_tag_index(Tag::when())->identifier());
  pair[1] = s->find(Tag::when());
  ABORT(!pair[1],"Tag::when not found");
  pairs[i] = V_Collection::make(Value::LIST, 2, pair);
  i++;
  ABORT(i != naddl, "Wrong number of ``additional'' tags");
  for (; i < len + naddl; i++) {
    pair[0] = String::make(QL_DD::find_by_tag_index(x[i-naddl])->identifier());
    pair[1] = s->find(x[i-naddl]);
    pairs[i] = V_Collection::make(Value::LIST, 2, pair);
  }
  ret = V_Collection::make(Value::LIST, len+naddl, pairs);
  delete pairs;
  return ret;
}


static int
find_tag(Tag a, Tag const tags[], int nelem)
{
  for (int i = 0; i < nelem; i++)
    if (tags[i] == a) return 1;
  return 0;
}


static Value *
m_tag_values(const QL_Term*, const Value *v)
{
  static size = 1024;
  static Tag tags[1024];
  static Value *values[1024], *list_v[1024];
  Step *s;
  int i, used = 0;

  const Material *m = v->down_to_MATERIAL_POINTER();
  Cursor_Step *c = m->step_cursor();
  for (i = 0, s = c->last(); s; s = c->previous(), i++) {
    if (!i) {
      tags[used] = Tag::step_kind();
      values[used] = s->find(Tag::step_kind());
      used++;
      tags[used] = Tag::when();
      values[used] = s->find(Tag::when());
      used++;
    }
      
    Tag *x = s->tag_array();
    int y = s->tag_number();
    for (int j = 0; j < y; j++) {
      if (!find_tag(x[j],tags,used)) {
	ABORT(used >= size, "Too many tags");
	tags[used] = x[j];
	values[used] = s->find(x[j]);
	used++;
      }
    }
  }
  delete c;

  Value *pair[2];
  for (i = 0; i < used; i++) {
    pair[0] = String::make(QL_DD::find_by_tag_index(tags[i])->identifier());
    pair[1] = values[i];
    list_v[i] = V_Collection::make(Value::LIST, 2, pair);
  }
  return V_Collection::make(Value::LIST, used, list_v);
}


BUILTIN_FIRST_DEFINITION(tag_value_bf_pred,1,1,one_undefined,1,1)
{
  if (m || n); // Suppress warnings.
  out_tv[0] = 0;
  switch (in_tv[0]->tv_type()) {
  case Value::MATERIAL_POINTER:
    out_tv[0] = m_tag_values(t,in_tv[0]);
    break;
  case Value::STEP_POINTER:
    out_tv[0] = s_tag_values(t,in_tv[0]);
    break;
  default:
    ql_eval_error(t,  "Argument 1 must of type 'MATERIAL_POINTER'",
		  " or 'STEP_POINTER' in");
    break;
  }
  return out_tv[0] ? (QL_Iterator_State*)1 : 0;
}


#include <sys/times.h>

static clock_t baseline_wall_time;
static struct tms baseline_other_times;

#if defined(GETRUSAGE)
#include <sys/time.h>
#include <sys/resource.h>
static struct rusage baseline_other_resources;
#endif


BUILTIN_FIRST_DEFINITION(baseline_rusage_pred,0,0,0,0,1)
{
  if (m || n || out_tv || in_tv || t); // Suppress warnings.
  baseline_wall_time = times(&baseline_other_times);
#if defined(GETRUSAGE)
  if (getrusage(RUSAGE_SELF,&baseline_other_resources) == -1) {
    ql_eval_error(t, "getrusage failed (", errno_msg(), ") in");
    return 0;
  }
#endif
  return (QL_Iterator_State*) 1;
}


#include <unistd.h>


BUILTIN_FIRST_DEFINITION(incremental_rusage_pred,0,0,0,1,1)
{
  if (m || n || in_tv); // Suppress warnings.

  static const long clk_tick = sysconf(_SC_CLK_TCK);

  if (clk_tick < 0) {
    ql_eval_error(t, "System error: sysconf produced an invalid CLK_TCK");
    return 0;
  }

  struct tms new_other_times;
  clock_t new_wall_time = times(&new_other_times);
  if (new_wall_time < baseline_wall_time) {
    ql_eval_error(t, "Wall clock time has overflowed");
    return 0;
  }
#if defined(GETRUSAGE)
  const int nelems = 3+9;
  // 3 time values plus idrss, minflt, majflt, swaps, inblock, outblock,
  // nsignals, nvcsw, nivcsw

  struct rusage new_other_resources;
  if (getrusage(RUSAGE_SELF,&new_other_resources) == -1) {
    ql_eval_error(t, "getrusage failed (", errno_msg(), ") in");
    return 0;
  }
#else
  const int nelems = 3;
#endif
  Value *outputs[nelems], *triple[3];

#define DELTA(I,FIELD,V1,V2) \
  triple[0] = String::make(MACRO_STRING(FIELD)); \
  if (errmsg()) { \
    ql_eval_error(t, errmsg(), " making an output 'STRING' in"); \
    return 0; \
  } \
  triple[1] = Float::make((V1 - V2 + 0.0) / clk_tick); \
  if (errmsg()) { \
    ql_eval_error(t, errmsg(), " making an output 'FLOAT' in"); \
    return 0; \
  } \
  triple[2] = Float::make((V1 + 0.0) / clk_tick); \
  if (errmsg()) { \
    ql_eval_error(t, errmsg(), " making an output 'FLOAT' in"); \
    return 0; \
  } \
  outputs[I] = V_Collection::make(Value::LIST,3,triple); \
  if (errmsg()) { \
    ql_eval_error(t, errmsg(), " making an output triple in"); \
    return 0; \
  }
  DELTA(0,u_cpu,new_other_times.tms_utime, baseline_other_times.tms_utime)
  DELTA(1,s_cpu,new_other_times.tms_stime, baseline_other_times.tms_stime)
  DELTA(2,elapsed, new_wall_time, baseline_wall_time)
#undef DELTA
#if defined(GETRUSAGE)
#define DELTA(I,FIELD) \
  triple[0] = String::make(MACRO_STRING(FIELD)); \
  if (errmsg()) { \
    ql_eval_error(t, errmsg(), " making an output 'STRING' in"); \
    return 0; \
  } \
  triple[1] = Long_Int::make((new_other_resources.CAT_ID2(ru_,FIELD) \
                  - baseline_other_resources.CAT_ID2(ru_,FIELD))); \
  if (errmsg()) { \
    ql_eval_error(t, errmsg(), " making an output 'INTEGER32' in"); \
    return 0; \
  } \
  triple[2] = Long_Int::make(new_other_resources.CAT_ID2(ru_,FIELD)); \
  if (errmsg()) { \
    ql_eval_error(t, errmsg(), " making an output 'INTEGER32' in"); \
    return 0; \
  } \
  outputs[I] = V_Collection::make(Value::LIST,3,triple); \
  if (errmsg()) { \
    ql_eval_error(t, errmsg(), " making an output triple in"); \
    return 0; \
  }
 
 DELTA(3,idrss);
  DELTA(4,minflt);
  DELTA(5,majflt);
  DELTA(6,nswap);
  DELTA(7,inblock);
  DELTA(8,oublock);
  DELTA(9,nsignals);
  DELTA(10,nvcsw);
  DELTA(11,nivcsw);
#undef DELTA
#endif

  out_tv[0] =  V_Collection::make(Value::LIST,nelems,outputs);
  if (errmsg()) {
    ql_eval_error(t, errmsg(), " making output list in");
    return 0;
  }

  return (QL_Iterator_State*) 1;

}


BUILTIN_FIRST_DEFINITION(db_size_pred,0,0,0,2,1)
{
  if (m || n || in_tv);
  off_t bytes;
  long blocks;
  if (QL_DD::db_size(&bytes, &blocks) != 0) {
    ql_eval_error(t, "Error getting database size (", errno_msg(), ") in");
    return 0;
  }
  out_tv[0] = Long_Int::make(bytes);
  if (errmsg()) {
    ql_eval_error(t, "Error making 'INTEGER32' for first return value (",
		  errmsg(), ") in");
    return 0;
  }
  out_tv[1] = Long_Int::make(blocks);
  if (errmsg()) {
    ql_eval_error(t, "Error making 'INTEGER32' for second return value (",
		  errmsg(), ") in");
    return 0;
  }
  return (QL_Iterator_State*) 1;
}

extern "C" {extern void InitVMTrace();}

BUILTIN_FIRST_DEFINITION(tracing_on_pred,0,0,0,1,1)
{
  if (m || n || in_tv);

  InitVMTrace();

  clear_error();
  out_tv[0] = String::make("Tracing is on");
  if (errmsg()) {
    ql_eval_error(t, "Error making 'String' for first return value (",
		  errmsg(), ") in");
    clear_error();
    return 0;
  }
  return (QL_Iterator_State*)1;
}

BUILTIN_FIRST_DEFINITION(tracing_off_pred,0,0,0,1,1)
{
  if (m || n || in_tv);

  InitVMTrace();

  clear_error();
  out_tv[0] = String::make("Tracing is off");
  if (errmsg()) {
    ql_eval_error(t, "Error making 'String' for first return value (",
                  errmsg(), ") in");
    clear_error();
    return 0;
  }
  return (QL_Iterator_State*)1;
}

