/* 
 * Copyright (c) 1995
 *       Whitehead Institute for Biomedical Research. All rights reserved.
 *
 * Released under terms specified in the Copyright file present in the top
 * directory of this distribution.
 */


#ifndef LIST_H
#define LIST_H

#ifndef BASICS_H
#include        "basics.H"
#endif

class Generic_Cursor;

class Generic_List
{
friend class Generic_Cursor;
public:
  Generic_List();
  Generic_List(int size);
  ~Generic_List() { delete _data; }

  Generic_Cursor *cursor_of() const;

  void put(int index, const void* element);
  void put(const void* element) { put (count(), element); }

  void* get(int index) const
  {
    if ((index < 0) || (index > high ())) 
      ABORT(1,"Index is out of range in Generic_List::get(int)"); 
    return _data [index];
  }
  
  // Return the index of element in `this' (based on pointer equality), or
  // low() - 1 if element is not found in `this'.
  int index(const void* element) const;

  // Remove element from the list (based on pointer equaility, and return it.
  // Return NULL if element is not in `this'.
  void* remove(const void* element);

  int contains(const void* element) const
  { return (index(element) >= low()); }

  int empty() const { return _count == 0; }
  int low() const { return 0; }
  int high() const { return _count - 1; }
  int count() const { return _count; }

protected:
  void* remove(int index);
  void** _data;
  void put_replace(int index, const void* element);
  void put_expand(int index);
  int _count;       // number of elements in the array
  int current_size; // amount of space allocated
};

class Generic_Cursor
{
public:
  void *first();
  void *next();
  void *last();
  void *previous();
  Generic_Cursor(const Generic_List *Arr):A(Arr),_cursor_index(0), _last_called(0){}
protected:
  int _cursor_index;
  int _last_called;  // a flag to ensure last() was called before previous()
  const Generic_List *A;
}; 


#define List(Type) CAT_ID2(List_,Type)
#define Cursor(Type) CAT_ID2(Cursor_,Type)


#define Listdeclare(Type) \
class List(Type); \
class Cursor(Type) : public Generic_Cursor \
{ \
public: \
  Type *first() { return (Type*) Generic_Cursor::first(); } \
  Type *next() { return (Type*) Generic_Cursor::next(); } \
  Type *last() { return (Type*) Generic_Cursor::last(); } \
  Type *previous() { return (Type*) Generic_Cursor::previous(); } \
  Cursor(Type) (const List(Type) *Arr) \
                         :Generic_Cursor( (const Generic_List *)Arr) { } \
}; \
class List(Type) : public Generic_List \
{ \
public: \
  List(Type) () : Generic_List () { } \
  List(Type) (int size) : Generic_List (size) { } \
  Cursor(Type) *cursor_of() const { \
	       return (Cursor(Type) *) Generic_List::cursor_of(); } \
  /* void copy (const List(Type)* src) */ \
  /*  { Generic_List::copy ((Generic_List*) src); } */ \
  void insert(const Type* element) { put(element); } \
  void put(const Type* element) { Generic_List::put ((void*) element); } \
  void put (int index, const Type* element) \
    { Generic_List::put (index, (void*) element); } \
  Type* get (int index) const \
    { return ((Type *)Generic_List::get (index)); } \
  Type* remove (int index) \
    { return ((Type *)Generic_List::remove (index)); } \
  Type* remove(const Type* element) \
    { return ((Type *)Generic_List::remove((void*) element)); } \
  int index (const Type* element) const \
    { return Generic_List::index ((void*) element); } \
  int contains(const Type* element) const \
    { return Generic_List::contains( (void*) element); } \
};
#endif		// last line of file
