/* 
 * Copyright (c) 1995
 *       Whitehead Institute for Biomedical Research. All rights reserved.
 *
 * Released under terms specified in the Copyright file present in the top
 * directory of this distribution.
 */


// $Id: main.C,v 1.6 1995/02/13 17:49:56 steve Exp $

#ifndef	__sys_types_h
#ifndef __TYPES_H
#include <sys/types.h> // For fstat
#endif
#endif

#ifndef	__sys_stat_h
#include <sys/stat.h>
#endif

#ifndef __SIGNAL_H
#include <signal.h>
#endif


#ifndef QL_DD_H
#include "ql_dd.H"
#endif

#ifndef BASICS_H
#include "basics.H"
#endif

#ifndef QL_ERROR_H
#include "ql_error.H"
#endif


extern int interactive_input, fifo_input;


int sigint_seen = 0;
static void sigint_handler(int)
{
  sigint_seen = 1;
}


static void sigterm_handler(int)
{
  ql_parse_error("SIGTERM seem; aborting uncommitted work and exiting");
  exit(0);
}


void main_start(int, char**);
void main_end(int, char**);

main(int argc, char** argv)
{
  ABORT(SIG_ERR == signal(SIGINT,sigint_handler),errno_msg());
  ABORT(SIG_ERR == signal(SIGTERM,sigterm_handler),errno_msg());

  main_start(argc, argv);
  
  QL_DD dd(argc == 1 ? 0 : argv[1]); 
  // argv[1] is the name of the database; if it is not specified, then operate
  // on an entirely transient database.
  
  // Figure out if we are getting input from a tty or a regular file:
  struct stat buf;
  int r = fstat(0,&buf);
  ABORT(r == -1, errno_msg());
  if (S_ISCHR(buf.st_mode)) interactive_input = 1;
  else if (S_ISFIFO(buf.st_mode)) fifo_input = 1;

  // Start parsing.
  extern int parse();
  parse();

  // Commit any remaining uncommitted changes.
  QL_DD::commit_last();

  (cout << "DONE\n").flush();
  main_end(argc, argv);
  exit(0);
}
