/* 
 * Copyright (c) 1995
 *       Whitehead Institute for Biomedical Research. All rights reserved.
 *
 * Released under terms specified in the Copyright file present in the top
 * directory of this distribution.
 */


#include "material.H"

#ifndef BASICS_H
#include "basics.H"
#endif

#ifndef VALUE_H
#include "value.H"
#endif

#ifndef QL_DD_ENTRY_H
#include "ql_dd_entry.H"
#endif

#ifndef QL_TERM_H
#include "ql_term.H"
#endif

#ifndef QL_ARGS_H
#include "ql_args.H"
#endif

#ifndef ID_INDEX_ENTRY_H
#include "id_index_entry.H"
#endif

#ifndef QL_DD_H
#include "ql_dd.H"
#endif

#ifndef STEP_H
#include "step.H"
#endif

int Material::expected_history_size = 20;


Material::Material(QL_DD_Material_Kind_Entry *dd_entry, char *id) :
Value(MATERIAL_POINTER), _dd_entry(dd_entry), _delete_marked(0)
{
  _history = PNEW(QL_DD::step_segment(),List_Step,(expected_history_size));
  _id = PNEW_ARRAY_ADDR(this,char,strlen(id) + 1);
  strcpy(_id,id);
}


// add_step should check for duplicate steps,
// and delete its argument if one is found.
// But it doesn't yet do that.
void Material::add_step(Step *s)
{
  ABORT(!s->timestamp(),"Need a timestamp");
  Cursor_Step *c = step_cursor();  
  // Cursor_Step *c = _history->cursor_of();

  int here = -1;
  int index = 0;

#ifdef FINISHED
Check that the when is not earlier than the creation of
the material.

Check for duplicates.

Unfortunately, the step has already been
created; furthermore, other materials may
use the step, and may already have had
their history updated to contain the step.

If we find a problem
we have three choices: we can try
a transaction rollback, we
can delete mark the step, or we
can remove it from all the histories
that contain it, and then delete it.
#endif

  Step *p = c->first();
  if (!p) _history->put(s); // Empty history.
  else {
    ABORT(s->timestamp() < p->timestamp(),
	"Cannot insert a step earlier than the create step");
    for ( ;p ;p = c->next(), index++)
      if ((here < 0) && (p->timestamp() > s->timestamp())) {
	here = index;
	break; // Maybe should continue to check for duplicates, but this
	       // can have some serious implications for paging
	       // activity on inserts.  Perhaps clustering (using os_segments)
	       // would help.
        // Code should also go here to check that no two steps have
	// the same timestamp.  This will require a command-line flag,
	// however, since logs translated from MapBase might have
	// more than one step with the same timestamp.
      }

    if (here < 0)
    _history->put(s); // No step with later time was found.
    else _history->put(here, s);
    // index is the index of the first step with a time later than that of s.
  }
  delete c;
}


void Material::print(ostream& os) const
{
  os <<  dd_entry()->identifier() << "("
    <<   id() << ")";
}


int Material::compare(const Value& tv) const
{
  if (tv.tv_type() == MATERIAL_POINTER)
    return tv.down_to_MATERIAL_POINTER() != this;
  else {
    set_error("Attempt to compare a Material_Value to a non-Material_Value");
    return 1;
  }
}


Material *Material::save(P_SEGMENT*)
{
  ABORT(1, "Do not call this function");
  return 0;
}

QL_FIRST_HEADER(Material::iterator_first_1_0)
{
  abort_m_n(m,1,n,0);
  ABORT(!in[0],"0 is illegal");
  if (in[0]->tv_type() != Value::MATERIAL_POINTER) {
    ql_eval_error(t, "Expect a value of type 'MATERIAL_POINTER' as the argument to");
    return 0;
  }
  ABORT(t->dd_entry()->entry_kind() != QL_DD_Entry::MATERIAL_KIND,
	"Need a material kind entry");
  QL_DD_Entry *e1 = in[0]->down_to_MATERIAL_POINTER()->dd_entry();
  QL_DD_Entry *e2 =  t->dd_entry();
  if(out); // To suppress warnings;
  return (QL_Iterator_State*) (e1 == e2);
}


QL_FIRST_HEADER(Material::iterator_first_0_1)
{
  abort_m_n(m,0,n,1);
  QL_DD_Material_Kind_Entry *m_k_entry 
    = t->dd_entry()->down_to_MATERIAL_KIND();
  // ID_INDEX_CURSOR is defined in Id_index_Entry.H.
  ID_INDEX_CURSOR *c  = m_k_entry->main_index()->ID_INDEX_CURSOR_OF();
  if (!c) { ql_eval_error(t, "Unable to obtain cursor"); return 0; }

  const Id_Index_Entry *index_entry = c->first();
  if (!index_entry) return 0;
  out[0] = index_entry->material();
  if (in); // Suppress warning.
  return (QL_Iterator_State *) c;
}


QL_NEXT_HEADER(Material::iterator_next_0_1)
{
  abort_m_n(m,0,n,1);
  if (in) t = 0; // Suppress compiler warnings.
  ID_INDEX_CURSOR *c = (ID_INDEX_CURSOR *)s;

  const Id_Index_Entry *index_entry = c->next();
  if (!index_entry) { delete c; return 0; }
  else {
    out[0] = index_entry->material();
    return s;
  }
}

Cursor_Step* Material::step_cursor() const
{
  return history()->cursor_of();
}


int Material::history_size() const { return history()->count(); }


time_t Material::creation_timestamp() const
{ return history()->get(0)->timestamp(); }
