/* 
 * Copyright (c) 1995
 *       Whitehead Institute for Biomedical Research. All rights reserved.
 *
 * Released under terms specified in the Copyright file present in the top
 * directory of this distribution.
 */


#ifndef QL_DD_H
#define QL_DD_H

// Needed for off_t:
#include <sys/types.h>

#ifndef BASICS_H
#include "basics.H"
#endif

#ifndef TAG_H
#include "tag.H"
#endif

#ifndef VALUE_H
#include "value.H"
#endif

class QL_DD_Entry;
class QL_DD_Tag_Entry;
class Value_Set_Table;

class QL_DD
{
public:
  // Create or open the database specified by `database'.  For ObjectStore and
  // Texas `database' is a file name.  If `database' is 0, work transiently.
  QL_DD(/* const -- Texas needs a non-const arg */ char *database);

  static QL_DD_Entry *find(char *);
  static QL_DD_Entry *find_or_error(char *);
  static QL_DD_Tag_Entry *tag_find(char *);
  // Return the dd entry for identifier, if any exists;
  // Otherwise return null.
  static QL_DD_Tag_Entry *find_by_tag_index(Tag);
  static void add_persistent(QL_DD_Entry *);  
  static P_DATABASE *db();
  static P_SEGMENT *transient_segment();
  static P_SEGMENT *default_segment();
  static P_SEGMENT *step_segment();
  static P_SEGMENT *tag_array_segment();
  static P_SEGMENT *material_segment();
  static void commit();  // Commit and start new xact if there is a database.
#ifdef OBJECTSTORE
  static void begin_nested();
  static void commit_nested();
  static void abort_nested();
#endif
  static void commit_last();  // Commit if there is a database.
  static void abort();  // Abort and start new xact if there is a database.
  static short get_and_increment_tag(); // Get the next tag index, and
                                        // increment it in the database.

  static Value_Set_Table *persistent_m_sets(); // The set of all persistent
						// material sets.
   
  // Function db_size returns 0 on success, -1 on failure.  On success,
  // *bytes contains the length of the database in bytes (as from stat),
  // and *blocks contains the actual number of blocks allocated.  For
  // ObjectStore and Texas these numbers do not include space for logs or
  // for schema (class layout) description.
  static int db_size(off_t *bytes, long* blocks);

  // Slot database_file is the name of the database (for ObjectStore and
  // Texas, this name is the path of the (primary) database file).
  SLOT(char*,database_file);

private:
  friend class DD_Ops;  // For operations that are private to QL_DD.
			// Putting them in a friend class
			// reduce the changes that affect clients
			// of QL_DD.

  QL_DD_Entry *_hardwireds;
  QL_DD_Entry *_persistents;
  P_DATABASE_ROOT *_dd_root, *_next_tag_root, *_material_set_root;
  P_TRANSACTION *_tx;
  P_DATABASE *_db;
  P_SEGMENT *_step_segment, *_tag_array_segment, *_material_segment, 
            *_default_segment, *_transient_segment;
  static QL_DD *_dd;
};

#endif
