/* 
 * Copyright (c) 1995
 *       Whitehead Institute for Biomedical Research. All rights reserved.
 *
 * Released under terms specified in the Copyright file present in the top
 * directory of this distribution.
 */


#include "ql_error.H"

#ifndef IOSTREAMH
#include <iostream.h>
#endif

#ifndef QL_TERM_H
#include "ql_term.H"
#endif

#ifndef QL_DD_ENTRY_H
#include "ql_dd_entry.H"
#endif

QL_Position error_position;

QL_Position QL_Position::operator += (int i)
{ _total += i; _column += i; return *this; }
void QL_Position::nl() { _column = 0; _line++; }

int ql_eval; // Initialized in ql.y::parse()
// ql_eval is set to 0 on error (either parse error or eval error) the function
// QL_eval checks ql_eval before evaluting a query term, and checks ql_eval
// afterward to determine if an error occured during evaluation.

static inline void all_finish(int line, int column)
{
  (cout << " near line " << line << ", column " << column << "\n").flush();
  ql_eval = 0;
}


static void ql_error_msg(const char *const m1, const char* const m2,
			 const char* const m3, const char* const m4,
			 const char* const m5, const char* const m6,
			 const char* const m7)
{
  cout << "ERROR=" << m1;
  if (m2) {
    cout << m2;
    if (m3) {
      cout << m3;
      if (m4) {
	cout << m4;
	if (m5) {
	  cout << m5;
	  if (m6) { cout << m6; if (m7) cout << m7; }
	}
      }
    }
  }
}


extern int yyleng; // Defined by the lexter;

void ql_parse_error(const char *m1, const char* m2, const char* m3,
		    const char* m4, const char* m5)
{
  ql_error_msg(m1, m2, m3, m4, m5, 0, 0);
  int column = error_position.column() - yyleng;
  column = column >= 0 ? column : 0;
  all_finish(error_position.line(), column);
}


void ql_eval_error(const QL_Term * const term, const char * const m1,
		   const char* const m2, const char* const m3,
		   const char* const m4, const char* const m5,
		   const char* const m6, const char* const m7)
{
  
  ql_error_msg(m1, m2, m3, m4, m5, m6, m7);
  cout << " predicate '" << term->dd_entry()->identifier() << "'";
  all_finish(term->parse_position().line(),term->parse_position().column());
}

