/* 
 * Copyright (c) 1995
 *       Whitehead Institute for Biomedical Research. All rights reserved.
 *
 * Released under terms specified in the Copyright file present in the top
 * directory of this distribution.
 */


// $Id: ql_iterator.C,v 1.2 1995/02/13 17:50:48 steve Exp $
#ifndef QL_ITERATOR_H
#include "ql_iterator.H"
#endif

QL_Iterator::QL_Iterator(QL_Term *t, int m, int n,
			 Builtin_First_Fn first_fn,
			 Builtin_Next_Fn next_fn,
			 int /* unbound_outputs_ok -- not used yet */)
: _term(t), _input_number(m), _output_number(n),
  _first(first_fn), _next(next_fn),
  _started(0), _state(0), _inputs(0), _outputs(0), _node_outputs(0)
{
  set_in_and_outputs(_term, _input_number, _output_number);
}


QL_Iterator::~QL_Iterator()
{
  clear_bindings(); // Unbinds any arguments bound during the iteration.
  if (_inputs) delete [] _inputs;
  if (_outputs) delete [] _outputs;
  if (_node_outputs) delete [] _node_outputs;
}


void QL_Iterator::set_in_and_outputs(QL_Term *t, int, int)
{
  _input_number=0, _output_number=0;
  QL_Args *p;
  for (p = t->args(); p; p = p->next())
    if (p->value()) _input_number++; else _output_number++;

  int in_index=0, out_index=0;
  if (_input_number) _inputs = new Value*[_input_number];
  if (_output_number) {
    _outputs = new Value*[_output_number];
    _node_outputs = new QL_Binding_Node*[_output_number];
  }

  for (p = t->args(); p; p = p->next()) {
    Value *ba = p->value();
    if (ba) {
      ABORT(in_index >= _input_number, "");
      _inputs[in_index] = ba;
      in_index++; 
    } else {
      ABORT(out_index >= _output_number, "");
      _node_outputs[out_index] = p->node();
      out_index++; 
    }
  }
}


void QL_Iterator::clear_bindings()
{
  for (int i=0; i < _output_number; i++)  {
    _node_outputs[i]->add_value(0);    
    _outputs[i] = 0;
  }
}


static int term_match(Value* source, QL_Binding_Node *target)
     // This fucntion will eventually implement Ullman's
     // "term matching" algorithm.  Right now it only handles
     // the simplest cases: either target is unbound, or bound
     // to a "complete" value (i.e. not a term with remaining
     // variabless).
{
  if (target->value()) {
    if (*source == *target->value()) return 1; else return 0;
  } 
  target->add_value(source);    
  return 1;
}


int QL_Iterator::set_bindings()
{
  for (int i=0; i < _output_number; i++) {
    // ABORT(!_outputs[i] && !unbound_outputs_ok,"No value output");
    if (!term_match(_outputs[i],_node_outputs[i])) return 0;
  }
  return 1;
}


extern int sigint_seen;  // Defined in main.C.

int QL_Iterator::true_bindings()
{
  for (;;) {

    if (sigint_seen) {  // Check for control-C.
      ql_eval_error(_term, "SIGINT seen while evaluating");
      ql_eval = 0;
      return 0;
    }
    
    clear_bindings();
    if (!_started){
      _started = 1;
      _state = _first(_term, _input_number, _inputs, _output_number, _outputs);
    } else {
      if (!_next) return 0;	// For command-like predicates, this is 
      // a convenience.
      _state
	= _next(_term,_state, _input_number, _inputs, _output_number, _outputs);
    }
    if (_state) {
      if (set_bindings()) return 1;
    }
    else return 0;
  }
}
