/* 
 * Copyright (c) 1995
 *       Whitehead Institute for Biomedical Research. All rights reserved.
 *
 * Released under terms specified in the Copyright file present in the top
 * directory of this distribution.
 */


// $Id: ql_term.H,v 1.4 1995/02/13 17:50:55 steve Exp $
#ifndef QL_TERM_H
#define QL_TERM_H

#ifndef __STDIO_H
#include <stdio.h>
#endif

#ifndef QL_ADORNMENT_H
#include "ql_adornment.H"
#endif

#ifndef QL_ERROR_H
#include "ql_error.H"
#endif

class QL_Builtin_Implementation;
class QL_Args;
class QL_DD_Entry;
class QL_DD_Tag_Entry;
class QL_Type_Descriptor;

#ifndef BASICS_H
#include "basics.H"
#endif

#ifndef VALUE_H
#include "value.H"
#endif

#ifndef QL_QC_H
#include "ql_gc.H"
#endif

class QL_Term : public Value
{
public:
  tv_type_id tv_type() const { return TERM; }
  QL_Term *save(P_SEGMENT*) const;
  void print(ostream&) const;
  int compare(const Value& a);

  virtual int is_tag_assoc() { return 0; }

  QL_Term(QL_DD_Entry *dd_entry, QL_Args *args, QL_Position position);

  void set_next(QL_Term *x) { _next = x; }

  static void *create(void *, void*, QL_Position&); // Used in the parser (ql.y) to
  				                   // provide a procedural abstraction
				                   // for much of the work of putting
				                   // together a parsed term.

  int set_implementation();
  // Set the implementation slot for term for a builtin
  // if it's not already set.
  // Returns 0 if no implementation can be found, otherwise 1;

  void set_adornment_info();
  // Sets the adornment, number_bound, and number_free
  // slots.


  SLOT(QL_Args*,args);
  SLOT(QL_DD_Entry*,dd_entry);
  SLOT(QL_Position,parse_position); // Not currently used.
  SLOT(QL_Term*,next);
  SLOT(int,term_position); // The 0-based index of the term in the query.
			   // Not yet used, but is designed to
			   // allow us to prohibit all but DBA's
			   // from issuing bulk updates on any but
			   // temporary material lists.

  // The following (for builtins) serve to
  // cache the results of lookup in the
  // adornment map:
  SLOT(QL_Builtin_Implementation*,implementation);
  SLOT(int,number_free);
  SLOT(int,number_bound);
  SLOT(QL_Adornment,adornment); // The adornment found at evaluation time.
  SLOT(QL_Adornment,used_adornment); // The adornment of the builtin
				     // that was used.  Use of underbound
				     // implementations is not yet
				     // supported.

protected:
  QL_GC_MEMBERS(QL_Term);
};


class QL_Tag_Assoc : public QL_Term
{
public:
  int is_tag_assoc() { return 1; }

  enum op_type {EQUAL, PLUS_EQUAL, MINUS_EQUAL};
  
  QL_Tag_Assoc(QL_DD_Entry*, QL_Args *args, QL_Position position);

  Value *check_or_coerce_type(QL_Term*);
  // Call ql_eval_error iff the argument is (or is bound to) a value
  // of the correct type.
  // In some cases, the return value is the argument value
  // coerced to the correct type.
  // Calls ql_eval_error and returns 0 if there is a type error.
  // The QL_Term* argument is needed for the error messages.

  SLOT(op_type,op);
  SLOT(QL_DD_Tag_Entry*,tag);
private:
  int check_type(QL_Term*, char*, QL_Args*, const QL_Type_Descriptor*[], int);
  Value *check_type_1_tag_value(QL_Term*, char*, Value*, const QL_Type_Descriptor*);
  Int *check_type_boolean(QL_Term*, char*, Value*);
  Int *check_type_integer(QL_Term*, char*, Value*);
  Value *check_type_dna_sequence(QL_Term*, char*, Value*);
  Value *check_type_collection(QL_Term*, char*, Value*, const QL_Type_Descriptor*);
};

extern int check_tag_value_type(QL_Term *,
				Value **, Value::tv_type_id, char *tag_id);
#endif
