/* 
 * Copyright (c) 1995
 *       Whitehead Institute for Biomedical Research. All rights reserved.
 *
 * Released under terms specified in the Copyright file present in the top
 * directory of this distribution.
 */


#ifndef STEP_H
#define STEP_H

#ifndef TAG_H
#include "tag.H"
#endif

#ifndef BASICS_H
#include "basics.H"
#endif

#ifndef VALUE_H
#include "value.H"
#endif

#ifndef LIST_H
#include "list.H"
#endif

class QL_DD_Tag_Entry;
class QL_DD_Step_Kind_Entry;
class QL_Tag_Assoc;
class Collated_Tags;
class Date;
class String;
class Int;
class QL_Args;

class List_Value;
class Step;
class List_Tag;

Listdeclare(Step);

class Step : public Value
{
public:

  Value *find(Tag tag) const; 
  // Return 0 if the Tag was not found,
  // otherwise the address of the first
  // Value* associated with the Tag.

  SLOT(char,delete_marked); // This order is necessary to
			    // conserve space.

  SLOT(short,tag_number);  // The number of tags in the step, _excluding_
                           // the ``when'' tag.

  SLOT(Tag*,tag_array);	   // All the non-``when'' tags in the step.  DO NOT
                           // MODIFY THE ELEMENTS of the return value.

  SLOT(QL_DD_Step_Kind_Entry*,step_kind);
  SLOT(time_t,timestamp);  // Contains the the value for the 'when' tag.

private:
  Value **_value;
  // Arrays of the tags and values in a step.
  // The number of values associated with a tag must be an integral
  // multiple of the number of values expected
  // to be associated with that tag.
  

  // void save_arg_in_database(QL_Args*); // Not needed?


public:
  void print(ostream&) const;
  Step *save(P_SEGMENT *dbs);  // Do not use!
  int compare(const Value&) const;

  Step(Collated_Tags *,QL_DD_Step_Kind_Entry*);
  // Argument 1 is a collection of Tag_Assoc values,
  // And must already contain the tag step_kind.

  void update(QL_Tag_Assoc *); // The Tag_Assoc indicates what kind of update(s?).
  //  Not implemented yet.

  void delete_mark() { _delete_marked = 1; }
  
  int contains_materials(List_Value*) const; 
  // Return 1 iff each Value in the argument is
  // the value of a tag in "this".  This function
  // does not examine Values that are elements
  // in a Collection Value--i.e. it only looks
  // at top-level Values.

  Date when() const; // Return the value of the when tag of this;

  // String *step_kind(); // Return the value of the step_kind tag of "this";

  // Int *step_id();  // Return the value of the step id of "this";

  int collides(Step *); // Return 1 if "this" collides with the argument;
			// otherwise return 0.  (Not yet implemented.)

  void add_to_materials(Collated_Tags*); // Add this to every material mentioned in 
					 // the Collated_Tags.
					 // Also update the indexes of any id tags.
					 // to point to (the--if there are any
					 // id tags) material.
					 // Error handling is not finished,
					 // though most user errors will
					 // have been checked for before calling
					 // this function.


};

#endif
